/*
 *   I use these routines just to decide when I have to fake a 
 *   volume-table to preserve compatability to original ftape.
 */
/*
 *      Copyright (C) 1994-1995 Bas Laarhoven.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 $Source: /homes/cvs/zftape/ftape-eof.c,v $
 $Author: claus $
 *
 $Revision: 1.6 $
 $Date: 1995/12/07 15:39:46 $
 $State: Exp $
 *
 *      This file contains the eof mark handling code
 *      for the QIC-40/80 floppy-tape driver for Linux.
 */

static char RCSid[] =
"$Id: ftape-eof.c,v 1.6 1995/12/07 15:39:46 claus Exp $";
  
  
#include "ftape.h"
#include <linux/string.h>
#include <linux/errno.h>


#include "ftape-eof.h"
#include "ftape-bsm.h"
#include "ftape-rw.h"

/*      Global vars.
 */

/*  a copy of the failed sector log from the header segment.
 */

/*
 *  the bad_sector_map is always of size 27kb because of the format 4 tapes.
 *  Thus we just take the upper part of the bsm to store the failed_sector_log.
 *
 *  Would it be nicer to define a "struct header_segment"? It would occupy hardly
 *  more memory than the current scheme.
 */
EofMark *eof_map = NULL;
 
/*  number of eof marks (entries in bad sector log) on tape.
 */
int nr_of_eof_marks = -1;


/*      Local vars.
 */

static char linux_tape_label[] = "Linux raw format V";
enum { min_fmt_version = 1, max_fmt_version = 2 };
static unsigned ftape_fmt_version = 0;


/*  Ftape (mis)uses the bad sector log to record end-of-file marks.
 *  Initially (when the tape is erased) all entries in the bad sector
 *  log are added to the tape's bad sector map. The bad sector log
 *  then is cleared.
 *
 *  The bad sector log normally contains entries of the form:
 *  even 16-bit word: segment number of bad sector
 *   odd 16-bit word: encoded date
 *  There can be a total of 448 entries (1792 bytes).
 *
 *  My guess is that no program is using this bad sector log (the
 *  format seems useless as there is no indication of the bad sector
 *  itself, only the segment)
 *  However, if any program does use the bad sector log, the format
 *  used by ftape will let the program think there are some bad sectors
 *  and no harm is done.
 *  
 *  The eof mark entries that ftape stores in the bad sector log:
 *  even 16-bit word: segment number of eof mark
 *   odd 16-bit word: sector number of eof mark [1..32]
 *  
 *  The eof_map as maintained is a sorted list of eof mark entries.
 *
 *
 *  The tape name field in the header segments is used to store a
 *  linux tape identification string and a version number.
 *  This way the tape can be recognized as a Linux raw format
 *  tape when using tools under other OS's.
 *
 *  'Wide' QIC tapes (format code 4) don't have a failed sector list
 *  anymore. That space is used for the (longer) bad sector map that
 *  now is a variable length list too.
 *  We now store our end-of-file marker list after the bad-sector-map
 *  on tape. The list is delimited by a (__u32) 0 entry.
 */

int
ftape_validate_label( char* label)
{
  TRACE_FUN( 8, "ftape_validate_label");
  int result = 0;

  TRACEx1( 4, "tape  label = `%s'", label);
  ftape_fmt_version = 0;
  if (memcmp( label, linux_tape_label, strlen( linux_tape_label)) == 0) {
    int pos = strlen( linux_tape_label);
    while (label[ pos] >= '0' && label[ pos] <= '9') {
      ftape_fmt_version *= 10;
      ftape_fmt_version = label[ pos++] - '0';
    }
    result = (ftape_fmt_version >= min_fmt_version &&
              ftape_fmt_version <= max_fmt_version);
  }
  TRACEx1( 4, "format version = %d", ftape_fmt_version);
  TRACE_EXIT;
  return result;
}

static EofMark*
find_end_of_eof_list( EofMark* ptr, byte* limit)
{
  while ((byte *)ptr + 3 < limit) {
    if ( ptr->entry ) {
      ++ptr;
    } else {
      return ptr;
    }
  }
  return NULL;
}

void
ftape_extract_file_marks( byte* address)
{
  TRACE_FUN( 8, "extract_file_marks");
  int i;

  eof_map = NULL;
  if (format_code == 4) {
    EofMark* end;
    EofMark* start = (EofMark *)find_end_of_bsm_list( address + 256,
                                       address + 29 * SECTOR_SIZE);
    nr_of_eof_marks = 0;
    if (start) {
      (byte *)start += 3;               /* skip end of list mark */
      end = find_end_of_eof_list( start, address + 29 * SECTOR_SIZE);
      if (end && ((byte *)end - (byte *)start) <= FAILED_SECTOR_LOG_SIZE ) {
	eof_map = (EofMark *)&bad_sector_map[   (SIZEOF(bad_sector_map) 
						 - ((byte *)end - (byte *)start)
						 + FAILED_SECTOR_LOG_SIZE-1)&~(sizeof(EofMark)-1) ];
	nr_of_eof_marks = (end - start) / sizeof( EofMark);
        memcpy( eof_map, start, nr_of_eof_marks * sizeof(EofMark) );
      } else {
        TRACE( 1, "File Mark List is too long or damaged !");
      }
    } else {
      TRACE( 1, "Bad Sector List is too long or damaged !");
    }
  } else {
    /*
     *   this routine gets called after the bsm already has been extracted
     */
    eof_map = (EofMark *)&bad_sector_map[ SIZEOF(bad_sector_map) - FAILED_SECTOR_LOG_SIZE ];
    memcpy( eof_map, address + 256, FAILED_SECTOR_LOG_SIZE );
    nr_of_eof_marks = GET2( address, 144);
  }
  TRACEi( 7, "number of file marks:", nr_of_eof_marks);
  if (ftape_fmt_version == 1) {
    TRACE( 2, "swapping version 1 fields");
    /*  version 1 format uses swapped sector and segment fields, correct that !
     */
    for (i = 0; i < nr_of_eof_marks; ++i) {
      __u16 tmp = eof_map[ i].mark.segment;
      eof_map[ i].mark.segment = eof_map[ i].mark.date;
      eof_map[ i].mark.date = tmp;
    }
  }
  for (i = 0; i < nr_of_eof_marks; ++i) {
    TRACEx2( 7, "eof mark: %5d/%2d",
            eof_map[ i].mark.segment, eof_map[ i].mark.date);
  }
  TRACE_EXIT;
}

/*
 *  this function simply eliminates the failed sector log by 
 *  setting all segments with failing sectors to zero.
 */
void
zftape_eliminate_failed_sector_log(  EofMark *eof_map, int nr_of_eof_marks )
{
  TRACE_FUN( 5, "zftape_eliminate_failed_sector_log");
  int i;
  unsigned failing_segment;

  for (i = 0; i < nr_of_eof_marks; ++i) {
    failing_segment = eof_map[ i].mark.segment;
    if ( !valid_segment_no( failing_segment ) ) {
      TRACEi( 2, "bad entry in failed sector log:", failing_segment);
    } else {
      /*
       *  well, the failed_sector_log is located right after the bad sectors 
       *  in the same buffer space as the bad_sector_map, but that should not
       *  collide as the bad_sector_map grows by at most 3 bytes whereas the
       *  failed sector log shrinks by 4 bytes.
       */
      put_bad_sector_entry( failing_segment, EMPTY_SEGMENT );
      bad_sector_map_changed= 1;
      TRACEx2( 7, "moved entry %d from failed sector log (%d)",
	      i, failing_segment);
    }
  }
  memset( eof_map, '\0', nr_of_eof_marks * sizeof(EofMark) );
  TRACE_EXIT;
}



